#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.



import sys, os
import random
from collections import namedtuple, OrderedDict
import tempfile

from PyQt4.QtCore import Qt, QTimer
from PyQt4.QtGui import QAction
from PyQt4.QtTest import QTest

from . import utils
from pybiklib.settings import settings


class Logger (object):
    logfilename = 'pybiktest/test.log'
    
    def __init__(self):
        self.lineno = None
        self.result = None
        self.logfile = None
        
    def open(self, log):
        if not log:
            self.log('Logfile disabled\n')
            return
        try:
            self.logfile = open(self.logfilename, 'wt', encoding='utf-8')
            self.logf('Use logfile: {}\n', self.logfilename)
        except IOError as e:
            self.logfile = None
            self.log('Logfile disabled:', e)
            
    def close(self):
        if self.logfile is not None:
            self.logfile.close()
            self.logfile = None
            
    def log(self, *args, **kwargs):
        if self.lineno is not None:
            args = ('line {}:'.format(self.lineno),) + args
        print(*args, **kwargs)
        if self.logfile is not None:
            kwargs['file'] = self.logfile
            print(*args, **kwargs)
            
    def logf(self, format, *args, **kwargs):
        self.log(format.format(*args), **kwargs)
        
    def error(self, message, *args, **kwargs):
        kwargs['file'] = sys.stderr
        self.log(message, *args, **kwargs)
        message = message.rstrip(':')
        if message not in self.result:
            self.result[message] = 1
        else:
            self.result[message] += 1
        
    def errorf(self, message, format, *args, **kwargs):
        self.error(message, format.format(*args), **kwargs)
        
logger = Logger()
log = logger.log
logf = logger.logf
log_error = logger.error
logf_error = logger.errorf


class NamedObject (object):     # pylint: disable=R0903
    def __init__(self, name):
        self.name = name
    def __repr__(self):
        return self.name
matchall = NamedObject('matchall')
nomatch = NamedObject('nomatch')


def mkState(fields):
    _State = namedtuple('_State', fields)
    class State(_State):   # pylint: disable=W0232
        def tostr(self, other=None):
            def field_tostr(i):
                if self[i] is matchall:
                    return ''
                elif other is None or self[i] != other[i]:
                    return '{}={!r}, '.format(self.fields[i], self[i])
                else:
                    return ' ' * (len(self.fields[i]) + len(repr(self[i])) + 3)
            return ''.join([field_tostr(i) for i in range(len(self))]).rstrip(', ')
            
        fields = _State._fields
        asdict = _State._asdict
        replace = _State._replace
    State.default = State(**{f: matchall for f in State.fields})
    return State
    
    
class StateInfo (object):   # pylint: disable=R0903
    def __init__(self):
        self.untested_tnames = None
        self.islimit = None
        
    def __str__(self):
        return 'StateInfo: untested_t: #{}, islimit: {}'.format(
                len(self.untested_tnames, self.islimit))
                
        
class StatesInfo (dict):
    def __init__(self):
        dict.__init__(self)
        
    def istested(self, *args):
        if len(args) == 0:
            for s in self.values():
                if s.untested_tnames and s.islimit:
                    return False
            return True
        elif len(args) == 1:
            state = args[0]
            return state in self and not (self[state].untested_tnames and self[state].islimit)
        else:
            raise TypeError('istested expected at most 1 argument, got %s' % len(args))
            
    def reset(self, untested_tnames):
        for v in self.values():
            v.untested_tnames = untested_tnames[:]
            
    def get_unreached(self):
        return [s for s, si in self.items() if si.untested_tnames and si.islimit]
        
        
class Transition (object):
    def __init__(self, func=None, exprs=None, states=None):
        self.func = func
        self.exprs = {} if exprs is None else exprs
        self.states = {} if states is None else states
        
        
class Transitions (dict):
    def __init__(self):
        self.limits = []
        self.stateinfos = StatesInfo()
        self.path_to_untested = []
        
    def islimit(self, state):
        state = state.asdict()
        for limit in self.limits:
            try:
                if not eval(limit, state):
                    return False
            except Exception as e:  # pylint: disable=W0703
                log_error('error in limit:', e)
        return True
        
    def reset(self):
        self.stateinfos.reset(list(self.keys()))
        
    def add_stateinfo(self, state):
        try:
            return self.stateinfos[state]
        except KeyError:
            stateinfo = dict.setdefault(self.stateinfos, state, StateInfo())
            stateinfo.untested_tnames = list(self.keys())
            stateinfo.islimit = self.islimit(state)
            return stateinfo
            
    def update_transition(self, transition, state, target):
        transition.states[state] = target
        self.add_stateinfo(state)
        if target is not None and target not in transition.states:
            transition.states[target] = None
            self.add_stateinfo(target)
            
    @staticmethod
    def _path_to_list(path):
        result = []
        while path is not None:
            item, path = path
            result.append(item)
        return result
        
    def get_path_to_untested(self, state):
        paths = [[(state, None), None]]
        reachable = set()
        while paths:
            path = paths.pop(0)
            state = path[0][0]
            stateinfo = self.stateinfos[state]
            reachable.add(state)
            if stateinfo.islimit and stateinfo.untested_tnames:
                return self._path_to_list(path)
            for name, transition in self.items():
                target = transition.states[state]
                if not stateinfo.islimit or self.stateinfos[target].islimit:
                    if target not in reachable:
                        paths.append([(target, name), path])
        return None
        
    def get_random_transition(self, state):
        if not self.path_to_untested:
            stateinfo = self.stateinfos[state]
            while stateinfo.untested_tnames:
                name = stateinfo.untested_tnames.pop(random.randrange(len(stateinfo.untested_tnames)))
                states = self[name].states
                target = states.get(state, None)
                if target is None or self.stateinfos[target].islimit:
                    return name
            self.path_to_untested = self.get_path_to_untested(state)
            if self.path_to_untested is None:
                raise Quit('all reachable states tested')
            target, name = self.path_to_untested.pop()
            assert name is None
            assert target == state
        target, name = self.path_to_untested.pop()
        assert name is not None
        return name
        
        
class Result (OrderedDict):
    __slots__ = ()
    
    def __init__(self):
        OrderedDict.__init__(self)
        self._dict = dict(visited=0, states=0, transitions=0)
        
    errors = property(lambda self: sum(self.values()))
    
    def __str__(self):
        summary = '\n  visited: {0.visited}, states: {0.states}, transitions: {0.transitions}'.format(self)
        errors = self.errors
        if errors:
            summary += '\n  errors: {}'.format(errors)
            for k, v in self.items():
                summary += '\n    {} {}'.format(v, k)
        return summary
        
    def __getattr__(self, key):
        if key.startswith('_'):
            raise AttributeError()
        return self._dict[key]
        
    def __setattr__(self, key, value):
        if key.startswith('_'):
            OrderedDict.__setattr__(self, key, value)
        else:
            self._dict[key] = value
            
    def __iadd__(self, other):
        for k, v in other._dict.items():    # pylint: disable=W0212
            self._dict[k] += v
        for k, v in other.items():
            self[k] = self.get(k, 0) + v
        return self
        
        
class Quit (Exception): pass    # pylint: disable=C0321


class TestRunner (object):
    _instance = None
    lineno = None
    
    def __init__(self, testdata_dir, test, test_args):
        assert self._instance is None
        self.current_test = test
        self.write_mode = 'no'
        self.log_widgets = False
        for a in test_args:
            if a.startswith('write-n'):
                self.write_mode = 'no'
            elif a.startswith('write-y'):
                self.write_mode = 'yes'
            elif a.startswith('write-e'):
                self.write_mode = 'error'
            elif a == 'log-widgets':
                self.log_widgets = True
        self.current_testfile = os.path.join(testdata_dir, test)
        self.oneway_transitions = []
        self.transitions = Transitions()
        self.widgets = {}
        self.conditions = []
        self.known_widget_functions = []
        self.field_functions = []
        
    def run(self, main_window):
        self.main_window = main_window
        self.running = True
        settings.draw.speed = 120
        QTimer.singleShot(0, self.loop)
        
    def isvalid(self, state):
        state = state.asdict()
        for condition in self.conditions[:]:
            try:
                if not eval(condition, state):
                    return False
            except Exception as e:  # pylint: disable=W0703
                log_error('error in condition:', e)
                self.conditions.remove(condition)
        return True
        
    def read_test(self):
        def parse_fields(line):
            fields = OrderedDict()
            if line != 'Fields:\n':
                for f in sorted(utils.field_functions.keys()):
                    fields[f] = []
                return parse_conditions, line, [fields]
            return parse_field, None, [fields]
        def parse_field(line, fields):
            if not line.startswith('  '):
                return parse_conditions, line, [fields]
            field, value = line.split('=', 1)
            field = field.strip()
            if field not in list(utils.field_functions.keys()):
                log_error('unknown field:', field)
            else:
                value = eval(value, {})
                fields[field] = value
            return parse_field, None, [fields]
        def parse_conditions(line, fields):
            self.State = mkState(fields)
            self.fields = fields
            if line != 'Conditions:\n':
                return parse_limits, line, []
            return parse_condition, None, []
        def parse_condition(line):
            if not line.startswith('  '):
                return parse_limits, line, []
            self.conditions.append(line.strip())
            return parse_condition, None, []
        def parse_limits(line):
            if line != 'Limits:\n':
                return parse_initialstate, line, []
            return parse_limit, None, []
        def parse_limit(line):
            if not line.startswith('  '):
                return parse_initialstate, line, []
            self.transitions.limits.append(line.strip())
            return parse_limit, None, []
        def _parse_state_value(default, line, exprs=None):
            try:
                state = eval('dict({})'.format(line), {})
            except SyntaxError:
                logf_error('error parsing state:', '{!r}', line)
                state = None
            else:
                exprs = list((exprs or {}).keys())
                for field in list(state.keys()):
                    if field not in self.State.fields:
                        log_error('unselected field:', field)
                        del state[field]
                    elif field in exprs:
                        logf_error('state contains expression field', '{}: {}', field, line)
                state = default and default.replace(**state)
            return state
        def parse_initialstate(line):
            if not line.startswith('Initial-State:'):
                self.initial_state = None
                return parse_transition, line, []
            line = line.split(':', 1)[1].strip()
            self.initial_state = _parse_state_value(self.State.default, line)
            return parse_oneway, None, []
        def parse_oneway(line):
            if not line.startswith('One-Way:'):
                return parse_transition, line, []
            name = line.split(':', 1)[1].strip()
            transition = Transition()
            self.oneway_transitions.append((name, transition))
            return parse_state, None, [transition, True]
        def parse_transition(line):
            if not line.startswith('Transition:'):
                return parse_unknown, line, ['transition']
            name = line.split(':', 1)[1].strip()
            transition = Transition()
            if name in self.transitions:
                logf_error('duplicate transition:', '{!r}', name)
            else:
                self.transitions[name] = transition
            return parse_field_expr, None, [transition]
        def parse_field_expr(line, transition):
            if not line.startswith('  Expression:'):
                return parse_state, line, [transition, False]
            line = line.split(':', 1)[1].strip()
            try:
                field, expr = line.split('=', 1)
            except ValueError:
                logf_error('error parsing expression:', '{!r}', line)
            else:
                field = field.strip()
                expr = expr.strip()
                if field in transition.exprs:
                    logf_error('duplicate expression', 'for {}: {!r}', field, expr)
                elif field not in self.State.fields:
                    log_error('unknown expression field:', field)
                else:
                    transition.exprs[field] = expr
            return parse_field_expr, None, [transition]
        def _replace_target(exprs, state, target, field, value):
            if field in exprs:
                try:
                    value = eval(exprs[field], state.asdict())
                except Exception as e:  # pylint: disable=W0703
                    logf_error('error parsing expression', '{!r}: {}', exprs[field], e)
            target_value = getattr(target, field)
            if target_value is matchall:
                if value is matchall:
                    value = getattr(state, field)
                target = target.replace(**{field: value})
            else:
                if field in exprs:
                    if value == target_value:
                        log_error('target contains expression field:', field, 'value:', target_value)
                    else:
                        log_error('target contains expression field:', field, 'value:', target_value,
                                    'expected:', value)
            return target
        def _expand_state(fields, state, target, exprs, i=0):
            if state is None or target is None:
                return
            if i >= len(fields):
                yield state, target
                return
            field, values = fields[i]
            if getattr(state, field) is matchall:
                for value in values:
                    rstate = state.replace(**{field: value})
                    for estate, etarget in _expand_state(fields, rstate, target, exprs, i+1):
                        etarget = _replace_target(exprs, estate, etarget, field, value)
                        yield estate, etarget
            else:
                for estate, etarget in _expand_state(fields, state, target, exprs, i+1):
                    etarget = _replace_target(exprs, estate, etarget, field, matchall)
                    yield estate, etarget
        def parse_state(line, transition, oneway):
            if not line.startswith('  State:'):
                return parse_oneway if oneway else parse_transition, line, []
            line = line.split(':', 1)[1].strip()
            state = _parse_state_value(self.State.default, line)
            return parse_target, None, [transition, state, oneway]
        def parse_target(line, transition, state, oneway):
            if not line.startswith('        '):
                target = self.State.default
            else:
                target = _parse_state_value(self.State.default, line.strip(), transition.exprs)
                line = None
            for state, target in _expand_state(list(self.fields.items()), state, target, transition.exprs):
                if state in transition.states and transition.states[state] is not None:
                    log_error('duplicate state:', state.tostr())
                elif self.isvalid(state):
                    if not self.isvalid(target):
                        log_error('invalid target state:', target.tostr())
                        target = None
                    if oneway:
                        transition.states[state] = target
                    else:
                        self.transitions.update_transition(transition, state, target)
            return parse_state, line, [transition, oneway]
        def parse_unknown(line, src):
            if line == 'End.' or src == 'end':
                return parse_unknown, None, ['end']
            logf_error('error parsing', '{}: {!r}', src, line)
            return parse_transition, None, []
            
        def process_line(func, line, *args):
            while line is not None:
                func, line, args = func(line, *args)
            return func, args
            
        try:
            func = parse_fields
            args = []
            with open(self.current_testfile, 'rt', encoding='utf-8') as testfile:
                for logger.lineno, line in enumerate(testfile):
                    if line.lstrip().startswith('#'):
                        continue
                    func, args = process_line(func, line, *args)
        finally:
            process_line(func, 'End.', *args)
            logger.lineno = None
                    
    def _ignore_fields(self, states, igntype, exprs=None):
        result = {}
        for i, field in enumerate(self.State.fields):
            target_value = nomatch
            for state, target in states.items():
                if igntype == 'unchanged':
                    if state[i] != target[i]:
                        result = {}
                        break
                    state = state.replace(**{field: matchall})
                    target = target.replace(**{field: matchall})
                elif igntype == 'sametarget':
                    if target_value in {nomatch, target[i]}:
                        state = state.replace(**{field: matchall})
                        target_value = target[i]
                    else:
                        result = {}
                        break
                elif igntype == 'expressions':
                    if field in exprs:
                        target = target.replace(**{field: matchall})
                else:
                    assert False, igntype
                if state in result and result[state] != target:
                    result = {}
                    break
                result[state] = target
            else:
                states = result
                result = {}
        return states
        
    def write_test(self):
        with open(self.current_testfile, 'wt', encoding='utf-8') as testfile:
            testfile.write('Fields:\n')
            for field, values in self.fields.items():
                print(' ', field, '=', values, file=testfile)
            testfile.write('Conditions:\n')
            for condition in self.conditions:
                print(' ', condition, file=testfile)
            testfile.write('Limits:\n')
            for limit in self.transitions.limits:
                print(' ', limit, file=testfile)
            testfile.write('Initial-State: %s\n' % self.initial_state.tostr())
            for name, transition in self.oneway_transitions:
                testfile.write('One-Way: %s\n' % name)
                states = list(transition.states.items())
                #assert len(states) == 1
                for state, target in states:
                    testfile.write('  State: {}\n'.format(state.tostr()))
                    if state != target and target.tostr(state).strip():
                        testfile.write('         {}\n'.format(target.tostr(state)))
            for name, transition in sorted(self.transitions.items()):
                testfile.write('Transition: %s\n' % name)
                states = {s:t for s, t in transition.states.items() if self.transitions.stateinfos[s].islimit}
                states = self._ignore_fields(states, 'expressions', transition.exprs)
                states = self._ignore_fields(states, 'unchanged')
                states = self._ignore_fields(states, 'sametarget')
                for field, expr in sorted(transition.exprs.items()):
                    testfile.write('  Expression: {} = {}\n'.format(field, expr))
                for state, target in sorted(states.items()):
                    testfile.write('  State: {}\n'.format(state.tostr()))
                    if state != target and target.tostr(state).strip():
                        testfile.write('         {}\n'.format(target.tostr(state)))
                    
    def get_state(self):
        state = {field: func() for field, func in self.field_functions}
        return self.State(**state)
        
    def find_qobjects(self, root_widget, write_template=False):
        def log_obj(indent, msg, name, obj):
            if self.log_widgets:
                logf('{}{}: {} ({})', '  '*indent, msg, name, obj.__class__.__name__)
            
        objects = [(root_widget, 0)]
        while objects:
            obj, indent = objects.pop()
            name = obj.objectName()
            if name in self.widgets:
                log_obj(indent, 'kwnobj', name, obj)
            elif name and not name.startswith('qt_'):
                self.widgets[name] = obj
                if isinstance(obj, QAction):
                    for tname, transition in self.oneway_transitions:
                        if tname == name:
                            assert transition.func is None
                            transition.func = obj.trigger
                    if name in self.transitions:
                        transition = self.transitions[name]
                        assert transition.func is None
                        transition.func = obj.trigger
                    elif write_template:
                        self.transitions[name] = Transition(func=obj.trigger)
                log_obj(indent, 'object', name, obj)
            else:
                log_obj(indent, 'intobj', name, obj)
            for child in reversed(obj.children()):
                objects.append((child, indent+1))
                
    def update_field_functions(self):
        self.field_functions = []
        mk_func = lambda wname, func: lambda: func(self.widgets[wname])
        mk_default = lambda default: lambda: default
        for field, (wname, default, func) in utils.field_functions.items():
            if field in self.State.fields:
                if wname in self.widgets:
                    self.field_functions.append((field, mk_func(wname, func)))
                else:
                    self.field_functions.append((field, mk_default(default)))
        
    def update_transition_functions(self, write_template=False):
        def mk_transition(func, widgets, *args):
            args = tuple(widgets) + args
            return lambda: func(*args)
            
        for wnames, fname, func in utils.widget_functions:
            if fname in self.known_widget_functions:
                continue
            if not all((n in self.widgets) for n in wnames):
                continue
            self.known_widget_functions.append(fname)
            widgets = [self.widgets[n] for n in wnames]
            for tname, transition in self.oneway_transitions + list(self.transitions.items()):
                if ' ' in tname:
                    tfname, args = tname.split(' ', 1)
                    if tfname != fname:
                        continue
                    try:
                        args = eval(args, {'Qt': Qt})
                    except Exception as e:  # pylint: disable=W0703
                        logf_error('error parsing expression', '{!r}: {}', args, e)
                        args = None
                    else:
                        if type(args) is not tuple:
                            args = (args,)
                else:
                    if tname != fname:
                        continue
                    args = ()
                assert transition.func is None, (wnames, fname, transition.func)
                if args is None:
                    transition.func = lambda: None
                else:
                    transition.func = mk_transition(func, widgets, *args)
            if write_template:
                assert fname not in self.transitions
                self.transitions[fname] = Transition(func=mk_transition(func, widgets))
                
    def init_test(self):
        try:
            self.read_test()
        except IOError as e:
            log('Error reading test data file:', e)
        assert list(self.fields.keys()) == list(self.State.fields), (list(self.fields.keys()), self.State.fields)
        write_template = False
        if not self.transitions:
            log_error('empty test')
            write_template = True
            
        # introspect ui
        self.find_qobjects(self.main_window, write_template)
        self.update_field_functions()
        self.update_transition_functions(write_template)
        
        # initial state
        self.current_state = self.get_state()
        if self.initial_state is None:
            self.initial_state = self.current_state
            log_error('missing initial state:', self.initial_state.tostr())
        elif self.initial_state != self.current_state:
            logf_error('wrong initial state:', '\n     found: {}\n  expected: {}',
                        self.current_state.tostr(), self.initial_state.tostr())
            self.initial_state = self.current_state
            
        self.transitions.reset()
        if write_template:
            raise Quit('template created')
        
    def check_transition(self, name, transition, oneway=False):
        if oneway and len(transition.states) > 1:
            logf_error('ambiguous ow transition:', '\n  {}: {} -> ({})',
                            name, self.current_state.tostr(), len(transition.states))
            for state in list(transition.states.keys()):
                if state != self.current_state:
                    del transition.states[state]
        target = transition.states.get(self.current_state, None)
        current_state = self.get_state()
        if target != current_state:
            if target is None:
                logf_error('unknown transition:', '\n  {}: {} -> {}',
                            name, self.current_state.tostr(), current_state.tostr())
            else:
                logf_error('wrong target', 'for {}: {}\n     found: -> {}\n  expected: -> {}',
                            name, self.current_state.tostr(), current_state.tostr(), target.tostr())
            if oneway:
                transition.states[self.current_state] = current_state
            else:
                self.transitions.update_transition(transition, self.current_state, current_state)
        self.current_state = current_state
        
    def check_result(self):
        for name, transition in list(self.transitions.items()):
            if transition.func is None:
                log_error('unused transition:', name)
                del self.transitions[name]
        all_states = list(self.transitions.stateinfos.keys())
        all_transitions = [(name, state) for name, transition in self.transitions.items()
                                            for state in transition.states.keys()]
        logger.result.states = len(all_states)    # pylint: disable=W0201
        logger.result.transitions = len(all_transitions)  # pylint: disable=W0201
        unreached = self.transitions.stateinfos.get_unreached()
        for state in unreached:
            log_error('unreached', state.tostr())
        field_values = {}
        for name, transition in self.transitions.items():
            states = transition.states
            for state, target in list(states.items()):
                if state in unreached or not self.transitions.stateinfos[state].islimit or target is None:
                    del states[state]
                    continue
                for field, svalue, tvalue in zip(self.State.fields, state, target):
                    field_values.setdefault(field, set()).update([svalue, tvalue])
        for field, values in self.fields.items():
            if field in field_values:
                if set(values) != field_values[field]:
                    log_error('changed field values:', field)
                    values[:] = sorted(field_values[field])
            else:
                log_error('unused field', field)
                
    def step(self, name, transition):
        postfunc = transition.func()
        if postfunc == 'find_qobjects':
            self.find_qobjects(self.main_window)
            self.update_field_functions()
            self.update_transition_functions()
        logger.result.visited += 1    # pylint: disable=E1101
        yield
        while self.main_window.is_animating():
            yield
            
    def loop(self):
        for unused in self._iloop():
            if not self.main_window.isVisible():
                log_error('Unexpected end of test')
                self.running = False
                return
            QTest.qWait(10)
        # without this line sometimes there is a segfault in the next window
        self.main_window.deleteLater()
            
    def _iloop(self):
        self.current_state = name = target = None  # for except statement
        try:
            log('Initializing test:', self.current_test)
            self.init_test()
            yield
            log('Running test:', self.current_test)
            for name, transition in self.oneway_transitions:
                #TODO python3.3: yield from self.step(name, transition)
                for unused in self.step(name, transition):
                    yield
                self.check_transition(name, transition, True)
            if self.current_state not in self.transitions.stateinfos:
                log_error('unknown state:', self.current_state)
                stateinfo = self.transitions.add_stateinfo(self.current_state)
                if not stateinfo.islimit:
                    log_error('state not in limit:', self.current_state)
            while True:
                name = self.transitions.get_random_transition(self.current_state)
                transition = self.transitions[name]
                #TODO python3.3: yield from self.step(name, transition)
                for unused in self.step(name, transition):
                    yield
                self.check_transition(name, transition)
        except Quit as e:
            self.check_result()
            if self.write_mode == 'yes' or logger.result.errors and self.write_mode == 'error':
                self.write_test()
                log('test data file written')
            logf('End of test {}: {}', self.current_test, e)
        except Exception:
            self.check_result()
            logf('exception in {}: {} -> {}', name,
                    self.current_state and self.current_state.tostr(),
                    target and target.tostr())
            sys.excepthook(*sys.exc_info())
        finally:
            if self.running:
                self.running = False
                self.main_window.close()
        
    @classmethod
    def wrap(cls, testdata_dir, tests, test_args):
        logger.open('log-file' in test_args)
        result = Result()
        cnt_tests = 0
        for test in tests:
            if '=' in test:
                test, repeat = test.split('=', 1)
                repeat = int(repeat)
            else:
                repeat = 1
            cnt_tests += repeat
            for unused in range(repeat):
                logger.result = Result()
                instance = cls(testdata_dir, test, test_args)
                fileno, instance.settings_file = tempfile.mkstemp(suffix='-settings.conf', prefix='pybiktest-', text=True)
                os.close(fileno)
                yield instance
                if instance.running:
                    log_error('Unexpected end of testrunner')
                log('Result:', logger.result)
                log('')
                os.remove(instance.settings_file)
                del instance.settings_file
                result += logger.result
        if cnt_tests > 1:
            logf('Summary ({}): {}', cnt_tests, result)
        logger.close()
        
    



#!/usr/bin/python3
#-*- coding:utf-8 -*-

#    Pybik -- A 3 dimensional magic cube game.
#    Copyright © 2009, 2011-2013  B. Clausius <barcc@gmx.de>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os, re

class Py2pyxParseError (Exception): pass


def main(argv):
    if not (1 < len(argv) <= 4):
        print('usage:', os.path.basename(__file__), 'python-file [pyx-filename [pxd-filename]]',
            file=sys.stderr)
        return 1
        
    arg_src = argv[1]
    arg_dst = argv[2] if len(argv) > 2 else None
    arg_pxd = argv[3] if len(argv) > 3 else None
    if not os.path.exists(arg_src):
        print('error:', arg_src, 'does not exist', file=sys.stderr)
        return 1
    create_pyx(arg_src, arg_dst, arg_pxd)
    return 0
    
def create_pyx(src_path, dst_path, pxd_path):
    errors = 0
    openr = lambda path: open(path, 'rt', encoding='utf-8')
    openw = lambda path: open(path, 'wt', encoding='utf-8') if path else sys.stdout
    edit_lines = None
    
    copy = lambda lineno, match: (match.string, None, None)
    add = lambda lineno, match: (match.expand(r'\1\4 \2\n'), None, None)
    def remove(unused_lineno, match):
        count = match.group(4)
        count = 1 if count == '' else int(count)
        return match.expand(r'\1#px \2\4\n'), None, [count, r'\1#px \2\4\n', None]
    replace = lambda lineno, match: (match.expand(r'\1\4 \2\n'), None, [1, r'\1#px \2\4\n', None])
    comment = lambda lineno, match: (match.expand(r'\1\2\4\n'), None, None)
    def repl_pxd(lineno, match):
        return (match.expand(r'\1\4 \2\n'),
                match.expand(r'\1\4').rstrip(':') + '    #line {}\n'.format(lineno),
                [1, r'\1#px \2\4\n', None])
    def move_pxd(unused_lineno, match):
        count = match.group(4)
        count = 1 if count == '' else int(count)
        return match.expand(r'\1#px \2\4\n'), None, [count, r'\1#px \2\4\n', r'\1\4\n']
    fdict = {None:copy, '+':add, '-':remove, '/':replace, ' ':comment,
             #'d ':pxd, 'd+':add_pxd,
             'd/':repl_pxd, 'd=':move_pxd}
    
    with openr(src_path) as srcf, openw(dst_path) as dstf, openw(pxd_path) as pxdf:
        pxdf.write('#file: {}\n\n'.format(src_path))
        for lineno, line in enumerate(srcf):
            try:
                match = re.match(r'^( *)(#px(d.|.)|)(.*?)\n$', line)
                if edit_lines is not None and edit_lines[0] > 0:
                    cnt_edit_lines, edit_re_pyxline, edit_re_pxdline = edit_lines
                    pyxline = edit_re_pyxline and match.expand(edit_re_pyxline)
                    pxdline = edit_re_pxdline and (match.expand(edit_re_pxdline).rstrip() + '    #line {}\n'.format(lineno))
                    edit_lines[0] -= 1
                    continue
                pyxline, pxdline, edit_lines = fdict[match.group(3)](lineno, match)
            except Exception as e:
                print('%s:%d:'%(src_path, lineno), e, file=sys.stderr)
                print('  Invalid line:', line, file=sys.stderr, end='')
                pyxline = line
                pxdline = None
                errors += 1
            finally:
                dstf.write(pyxline)
                if pxdline is not None:
                    pxdf.write(pxdline)
    if errors:
        raise Py2pyxParseError('create_pyx failed with %s errors' % errors)
        
    
if __name__ == '__main__':
    sys.exit(main(sys.argv))
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2009, 2011-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=W0201


from glob import glob
import os
import sys
import re

import distutils

from distutils.core import setup
import distutils.cmd
from distutils.extension import Extension

from distutils.dir_util import remove_tree
from distutils.file_util import move_file
from distutils.dep_util import newer, newer_group
from distutils.log import warn, info, error, debug

import distutils.dist

import distutils.command.build
import distutils.command.build_ext
import distutils.command.build_scripts
import distutils.command.install
import distutils.command.install_data
import distutils.command.install_lib
import distutils.command.clean
import distutils.command.sdist

#from Cython.Distutils import build_ext

from tools.po_isempty import po_isempty
from pybiklib import config


class Distribution (distutils.dist.Distribution):
    def __init__(self, attrs=None):
        self.bug_contact = None
        self.qt_ui_files = None
        distutils.dist.Distribution.__init__(self, attrs=attrs)


class build(distutils.command.build.build):
    """Adds extra commands to the build target."""
    user_options = distutils.command.build.build.user_options + [
            ('inplace', 'i', 'ignore build-lib and put compiled modules into the source directory'),
            ('use-generated-code', None, 'Use generated Code included in distributed source'
                                         ' to avoid build-dependencies'),
            ('parallel=', None, 'Build in parallel with the given number of processes or "auto"'),
            ('arch-only', None, 'Build only architecture dependent files'),
            ('indep-only', None, 'Build only architecture independent files'),
            ('fast', None, 'Build models without optimization'),
        ]
    boolean_options = distutils.command.build.build.boolean_options + [
            'inplace', 'use-generated-code', 'arch-only', 'indep-only', 'fast',
        ]
        
    def has_pure_modules(self):
        return self.indep_only and not self.inplace and distutils.command.build.build.has_pure_modules(self)
        
    _commands_dict = dict(distutils.command.build.build.sub_commands)
    # pylint: disable=W0212
    sub_commands = [('build_py',      has_pure_modules),
                    ('build_ext',     lambda self: self.arch_only and self._commands_dict['build_ext'](self)),
                    ('build_scripts', lambda self: self.indep_only and self._commands_dict['build_scripts'](self)),
                    ('build_ui',      lambda self: self.indep_only),
                    ('build_models',  lambda self: self.indep_only),
                    ('build_i18n',    lambda self: self.indep_only),
                    ('build_man',     lambda self: self.indep_only and not self.inplace),
                ]
    # pylint: enable=W0212
                
    def initialize_options(self):
        distutils.command.build.build.initialize_options(self)
        self.inplace = 0
        self.use_generated_code = False
        self.parallel = None
        self.arch_only = False
        self.indep_only = False
        self.fast = False
        
    def finalize_options(self):
        distutils.command.build.build.finalize_options(self)
        if not self.arch_only and not self.indep_only:
            self.arch_only = self.indep_only = True
        
        
class _build_csrc (distutils.command.build_ext.build_ext):
    user_options = distutils.command.build_ext.build_ext.user_options + [
        ('cython-opts=', None,          'Cython options'),
        ('gles=',        None,          'Build for GLES (experimental)'),
    ]
    
    def initialize_options(self):
        distutils.command.build_ext.build_ext.initialize_options(self)
        self.inplace = None
        self.cython_opts = ''
        self.gles = None
        
    def finalize_options(self):
        distutils.command.build_ext.build_ext.finalize_options(self)
        self.set_undefined_options('build',
                                       ('inplace', 'inplace'),
                                    )
        self.cython_opts = self.cython_opts.split()
        
    def compare_file(self, filename1, filename2):
        if self.force or not os.path.exists(filename2):
            return False
        with open(filename1, 'rb') as file1, open(filename2, 'rb') as file2:
            return file1.read() == file2.read()
            
    @staticmethod
    def run_py2pyx(py_file, pyx_file, pxd_file):
        info('py2pyx: %s --> %s, %s' % (py_file, pyx_file, pxd_file))
        from tools.py2pyx import create_pyx, Py2pyxParseError
        try:
            create_pyx(py_file, pyx_file, pxd_file)
        except Py2pyxParseError as e:
            error("error: %s", e)
            sys.exit(1)
        
    def run_cython(self, cython_opts, infile, outfile):
        info('cython: %s --> %s' % (infile, outfile))
        self.spawn(['cython3', '-3'] + cython_opts + ['-o', outfile, infile])
        
    def build_pyx(self, py_file, pyx_file, pxd_file):
        pxd_file_tmp = pxd_file + '.tmp'
        self.mkpath(os.path.dirname(pyx_file))
        self.make_file([py_file], pyx_file, self.run_py2pyx, (py_file, pyx_file, pxd_file_tmp))
        if os.path.exists(pxd_file_tmp):
            if self.compare_file(pxd_file_tmp, pxd_file):
                os.remove(pxd_file_tmp)
                info("unchanged file '%s'", pxd_file)
            else:
                if os.path.exists(pxd_file):
                    os.remove(pxd_file)
                move_file(pxd_file_tmp, pxd_file)
        
    def build_extension(self, extension):
        sources = extension.sources
        depends = sources + extension.depends
            
        # generate pxd- and pyx-files from py-files with #px annotations
        pyx_files = []
        pyx_files_dep = []
        for in_file in depends:
            base, ext = os.path.splitext(in_file)
            if ext == '.py':
                base = os.path.join(self.build_temp, os.path.splitext(in_file)[0])
                basename = os.path.basename(base)
                dirname = os.path.dirname(base)
                pyx_file = os.path.join(dirname, '_%s.pyx' % basename)
                pxd_file = os.path.join(dirname, '_%s.pxd' % basename)
                self.build_pyx(in_file, pyx_file, pxd_file)
                if in_file in extension.sources:
                    pyx_files.append(pyx_file)
                pyx_files_dep.append(pxd_file)
                
                if self.gles is not None:
                    info('Changing source for GLES %s', self.gles)
                    with open(pyx_file, 'rt', encoding='utf-8') as f:
                        text = f.read()
                    text = text.replace("DEF SOURCEGLVERSION = 'GL'", "DEF SOURCEGLVERSION = 'GLES%s'" % self.gles)
                    with open(pyx_file, 'wt', encoding='utf-8') as f:
                        f.write(text)
            elif ext in ('.pxi','.pxd'):
                out_file = os.path.join(self.build_temp, in_file)
                self.mkpath(os.path.dirname(out_file))
                self.copy_file(in_file, out_file)
                pyx_files_dep.append(out_file)
            else:
                if in_file in extension.sources:
                    pyx_files.append(in_file)
                else:
                    pyx_files_dep.append(in_file)
                    
        # generate C-files with cython
        if extension.language == 'c++':
            c_file_ext = '.cpp'
            cython_opts = ['--cplus'] + self.cython_opts
        else:
            c_file_ext = '.c'
            cython_opts = self.cython_opts
        c_files = []
        for in_file in pyx_files:
            base, ext = os.path.splitext(in_file)
            if ext == '.pyx':
                out_file = base + c_file_ext
                inplace_file = os.path.join('csrc', os.path.basename(out_file))
                if self.inplace and os.path.exists(inplace_file) and not self.force:
                    self.mkpath(os.path.dirname(out_file))
                    self.copy_file(inplace_file, out_file)
                else:
                    self.make_file(pyx_files+pyx_files_dep, out_file,
                                   self.run_cython, (cython_opts, in_file, out_file))
                c_files.append(out_file)
            else:
                c_files.append(in_file)
                
        # exclude Cython-files from dependencies
        c_files_dep = []
        for in_file in pyx_files_dep:
            base, ext = os.path.splitext(in_file)
            if ext not in ('.pxi', '.pxd'):
                c_files_dep.append(in_file)
        extension.depends = c_files_dep
        extension.sources = c_files
        
        
class build_csrc (_build_csrc):
    description = "create C source code for Python extensions"
    
    def build_extension(self, extension):
        _build_csrc.build_extension(self, extension)
        for c_file in extension.sources:
            inplace_file = os.path.join('csrc', os.path.basename(c_file))
            self.mkpath('csrc')
            self.copy_file(c_file, inplace_file)
            
        
class build_ext (_build_csrc):
    def build_extension(self, extension):
        if self.gles == '2':
            extension.libraries[:] = ['GLESv2']
        sources = extension.sources
        ext_path = self.get_ext_fullpath(extension.name)
        depends = sources + extension.depends
        if self.force or newer_group(depends, ext_path, 'newer'):
            info('building C-Code for %r extension', extension.name)
        else:
            debug('skipping %r extension (up-to-date)', extension.name)
            return
            
        # build C-code
        _build_csrc.build_extension(self, extension)
        # build extension module from C-code
        distutils.command.build_ext.build_ext.build_extension(self, extension)
        #HACK: Due to _build_csrc.compare_file the C compiler may not run, even though
        #      the dependencies expect this. Therefore update the timestamp manually.
        ext_path = self.get_ext_fullpath(extension.name)
        try:
            os.utime(ext_path, None)
        except OSError:
            pass
        
        
class build_scripts (distutils.command.build_scripts.build_scripts):
    user_options = distutils.command.build_scripts.build_scripts.user_options + [
            ('inplace',     'i', 'ignore build-lib and put compiled modules into the source directory'),
            ('build-temp=', 't', "temporary build directory"),
        ]
    boolean_options = distutils.command.build.build.boolean_options + ['inplace']
    
    def initialize_options(self):
        distutils.command.build_scripts.build_scripts.initialize_options(self)
        self.inplace = None
        self.build_temp = None
        
    def finalize_options(self):
        distutils.command.build_scripts.build_scripts.finalize_options(self)
        self.set_undefined_options('build',
                    ('inplace', 'inplace'),
                    ('build_temp', 'build_temp'),
                )
                
    def run(self):
        build_dir = self.build_dir
        self.build_dir = self.build_temp
        distutils.command.build_scripts.build_scripts.run(self)
        self.build_dir = build_dir
        for script in self.scripts:
            outfile = os.path.basename(script)
            script = os.path.join(self.build_temp, outfile)
            script = os.path.join(self.build_temp, outfile)
            if not os.path.exists(script):
                continue
            if not self.inplace:
                self.mkpath(self.build_dir)
                outfile = os.path.join(self.build_dir, outfile)
            outfile, ext = os.path.splitext(outfile)
            if ext != '.py':
                outfile += ext
            self.copy_file(script, outfile)
        
        
class build_ui(distutils.cmd.Command):
    description = "Compile Qt user interfaces files (.ui) and resources to python modules."
    
    user_options = [
        ('indent=',     'I', 'set indent width to N spaces, tab if N is 0 (default: 4)'),
        ('ui-execute',  'x', 'generate extra code to test and display the class'),
        ('from-imports','F', 'generate imports relative to "."'),
        ('build-lib=',  'b', 'directory for compiled UI modules'),
        ('inplace',     'i', 'ignore build-lib and put compiled UI modules into the source '
                             'directory alongside your pure Python modules'),
        ('force',       'f', 'forcibly build everything (ignore file timestamps)'),
        ('use-generated-code', None, 'Use generated Code included in distributed source'
                                     ' to avoid build-dependencies'),
    ]
    boolean_options = ['from-imports', 'ui-execute', 'inplace', 'force', 'use-generated-code',
                       ]
    
    def initialize_options(self):
        self.qt_ui_files = None
        self.indent = 4
        self.ui_execute = False
        self.from_imports = False
        self.build_lib = None
        self.inplace = None
        self.force = None
        self.use_generated_code = None
        
    def finalize_options(self):
        self.qt_ui_files = self.distribution.qt_ui_files
        self.set_undefined_options('build',
                                       ('build_lib', 'build_lib'),
                                       ('inplace', 'inplace'),
                                       ('force', 'force'),
                                       ('use_generated_code', 'use_generated_code'),
                                   )
                
    def run(self):
        for package_dir, ui_files in self.qt_ui_files:
            if self.use_generated_code:
                if not self.inplace:
                    self.copy_tree(package_dir, os.path.join(self.build_lib, package_dir))
            else:
                if not self.inplace:
                    package_dir = os.path.join(self.build_lib, package_dir)
                self.mkpath(package_dir)
                init_file = os.path.join(package_dir, '__init__.py')
                if self.force and os.path.exists(init_file):
                    os.remove(init_file)
                if not os.path.exists(init_file):
                    self.execute(lambda: open(init_file, 'wt', encoding='utf-8').close(),
                                 [], 'generating ' + init_file)
                for ui_file in ui_files:
                    filebase = os.path.splitext(os.path.basename(ui_file))[0]
                    py_file = os.path.join(package_dir, filebase + '.py')
                    self.make_file(ui_file, py_file, self.compile_ui, [ui_file, py_file])
                    
    def compile_ui(self, ui_file, py_file):
        from tools import translation_from_ui
        windowinfo = translation_from_ui.parse_file(ui_file)
        with open(py_file, 'wt', encoding='utf-8') as file:
            translation_from_ui.write_file(*windowinfo, file=file)
            

class build_models(distutils.cmd.Command):
    description = "Create data for Pybik models and write it to python modules."
    
    user_options = [
        ('inplace',     'i', 'ignore build-lib and put compiled UI modules into the source '
                             'directory alongside your pure Python modules'),
        ('force',       'f', 'forcibly build everything (ignore file timestamps)'),
        ('parallel=',   None,'Build in parallel with the given number of processes or "auto"'),
        ('fast',        None,'Build models without optimization'),
    ]
    boolean_options = ['inplace', 'force', 'fast']
    
    def initialize_options(self):
        self.build_base = None
        self.inplace = None
        self.force = None
        self.parallel = None
        self.fast = None
        
    def finalize_options(self):
        self.set_undefined_options('build',
                                       ('build_base', 'build_base'),
                                       ('inplace', 'inplace'),
                                       ('force', 'force'),
                                       ('parallel', 'parallel'),
                                       ('fast', 'fast'),
                                   )
        if self.parallel is None:
            self.parallel = 1
        elif self.parallel == 'auto':
            import multiprocessing
            self.parallel = multiprocessing.cpu_count()
        else:
            self.parallel = int(self.parallel)
                
    def run(self):
        from tools import modeldata
        modeldir = os.path.join('data' if self.inplace else self.build_base, 'models')
        self.mkpath(modeldir)
        if self.force and not self.dry_run:
            for filename in os.listdir(modeldir):
                filename = os.path.join(modeldir, filename)
                os.remove(filename)
        def create_modeldata(directory, testfunc):
            modeldata.create_modeldata(directory, testfunc, self.parallel, self.fast)
        if self.force:
            testfunc = None
        else:
            testfunc = lambda filename: newer('pybiklib/model.py', filename)
        message = 'generating model data using {} processes'.format(self.parallel)
        self.execute(create_modeldata, [modeldir, testfunc], msg=message)
        data_files = self.distribution.data_files
        for filename in os.listdir(modeldir):
            sourcepath = os.path.join(modeldir, filename)
            data_files.append(('share/pybik/models', (sourcepath,)))
        
        
class build_i18n(distutils.cmd.Command):
    description = "integrate the gettext framework"
    
    user_options = [
        ('desktop-files=',  None, '.desktop.in files that should be merged'),
        ('xml-files=',      None, '.xml.in files that should be merged'),
        ('schemas-files=',  None, '.schemas.in files that should be merged'),
        ('key-files=',      None, '.key.in files that should be merged'),
        ('domain=',         'd',  'gettext domain'),
        ('merge-po',        'm',  'merge po files against template'),
        ('po-dir=',         'p',  'directory that holds the i18n files'),
        ('inplace',         'i',  'ignore build-lib and put compiled UI modules into the source '
                                  'directory alongside your pure Python modules'),
        ('force',           'f',  'forcibly build everything (ignore file timestamps)'),
    ]
        
    boolean_options = ['merge-po', 'inplace', 'force']
    
    def initialize_options(self):
        self.build_base = None
        self.desktop_files = []
        self.xml_files = []
        self.key_files = []
        self.schemas_files = []
        self.domain = None
        self.merge_po = False
        self.po_dir = None
        self.inplace = None
        self.force = None
        
    def finalize_options(self):
        self.set_undefined_options('build', ('build_base', 'build_base'),
                                            ('inplace', 'inplace'),
                                            ('force', 'force'),
                                  )
        if self.inplace:
            self.mo_dir = 'data/locale'
        else:
            self.mo_dir = os.path.join(self.build_base, 'mo')
        if self.domain is None:
            self.domain = self.distribution.metadata.name
        if self.po_dir is None:
            self.po_dir = "po"
            
    def do_merge_po(self):
        '''Update po(t) files and print a report'''
        
        from tools.conv_plugin_for_translation import convert
        # We have to change the working dir to the po dir for intltool
        wd = os.getcwd()
        temp_files = []
        try:
            for unused_target, files in self.distribution.data_files:
                for in_file in files:
                    if in_file.endswith('.algorithm'):
                        out_file = in_file + '.py'
                        self.make_file(in_file, out_file, convert, [in_file, out_file])
                        temp_files.append(out_file)
            os.chdir(self.po_dir)
            self.spawn(["intltool-update", "-r", "-g", self.domain])
        finally:
            os.chdir(wd)
            for temp_file in temp_files:
                os.remove(temp_file)
                
    def run(self):
        """Update the language files, generate mo files and add them to the to be installed files"""
        
        os.environ["XGETTEXT_ARGS"] = "--msgid-bugs-address=%s " % self.distribution.bug_contact
            
        # Print a warning if there is a Makefile that would overwrite our values
        if os.path.exists("%s/Makefile" % self.po_dir):
            self.announce(''
                    'WARNING: Intltool will use the values specified from the'
                    '         existing po/Makefile in favor of the vaules'
                    '         from setup.cfg.'
                    '         Remove the Makefile to avoid problems.')
                    
        if self.merge_po:
            #TODO: --merge-po should do this and nothing else
            self.do_merge_po()
            
        data_files = self.distribution.data_files
        for po_file in glob("%s/*.po" % self.po_dir):
            if po_isempty(po_file):
                print('skipping empty po file', po_file)
                continue
            lang = os.path.splitext(os.path.basename(po_file))[0]
            mo_dir =  os.path.join(self.mo_dir, lang, "LC_MESSAGES")
            mo_file = os.path.join(mo_dir, "%s.mo" % self.domain)
            self.mkpath(mo_dir)
            def msgfmt(po_file, mo_file):
                self.spawn(["msgfmt", po_file, "-o", mo_file])
            self.make_file([po_file], mo_file, msgfmt, [po_file, mo_file])
            
            targetpath = os.path.join("share/locale", lang, "LC_MESSAGES")
            data_files.append((targetpath, (mo_file,)))
            
            
        if not self.merge_po:
            return
        #FIXME: currently *.desktop is treated as source and generated from *.desktop.in with merge-po.sh
        #       better: *.desktop.in is treated as source and *.desktop not committed
        # merge .in with translation
        if not self.desktop_files:
            for targetpath, files in data_files:
                for f in files:
                    if f.endswith('.desktop'):
                        self.desktop_files.append(f)
        for (files, switch) in ((self.xml_files, "-x"),
                                 (self.desktop_files, "-d"),
                                 (self.schemas_files, "-s"),
                                 (self.key_files, "-k"),):
            for f in files:
                if f.endswith(".in"):
                    self.warn("aDon't know, what to do with this file: " + f)
                    continue
                srcfile = f + '.in'
                if not os.path.isfile(srcfile):
                    self.warn("sDon't know, what to do with this file: " + f)
                    continue
                dstfile = f
                def intltool_merge(switch, src, dst):
                    self.spawn(["intltool-merge", switch, self.po_dir, src, dst])
                self.make_file(glob('%s/*.po'%self.po_dir)+[srcfile], dstfile,
                               intltool_merge, [switch, srcfile, dstfile])
                
    
class build_man (distutils.cmd.Command):
    description = "build the manpage"
    user_options = [
        ('force',       'f',  'forcibly build everything (ignore file timestamps)'),
    ]
    
    def initialize_options(self):
        self.force = None
        self.build_base = None
        self.build_dir = None
        
    def finalize_options(self):
        self.set_undefined_options('build', ('build_base', 'build_base'),
                                            ('force', 'force'),
                                  )
        self.set_undefined_options('build_scripts', ('build_dir', 'build_dir'))
        
    def run(self):
        def create_manpage(script_file, section, man_file):
            os.environ['PYTHONPATH'] = '.'
            self.spawn(['help2man', '--name', config.SHORT_DESCRIPTION, '--section', section,
                                    '--output', man_file, '--no-info', script_file])
            
        data_files = self.distribution.data_files
        man_dir = os.path.join(self.build_base, 'man')
        self.mkpath(man_dir)
        if '.' not in sys.path:
            sys.path.insert(0, '.')
        section = '6'
        for script in os.listdir(self.build_dir):
            script_file = os.path.join(self.build_dir, script)
            man_file = os.path.join(man_dir, '.'.join((script, section)))
            self.make_file(['pybiklib/config.py', script_file], man_file,
                           create_manpage, [script_file, section, man_file])
            data_files.append(('share/man/man'+section, [man_file]))
            
            
class install_lib (distutils.command.install_lib.install_lib):
    user_options = distutils.command.install_lib.install_lib.user_options + [
            ('arch-only', None, 'Install only architecture dependent files'),
            ('indep-only', None, 'Install only architecture independent files'),
        ]
    boolean_options = ['arch-only', 'indep-only']
    
    def initialize_options(self):
        distutils.command.install_lib.install_lib.initialize_options(self)
        self.data_dir = None
        self.arch_only = None
        self.indep_only = None
        
    def finalize_options(self):
        distutils.command.install_lib.install_lib.finalize_options(self)
        self.set_undefined_options('install',
                                       ('arch_only', 'arch_only'),
                                       ('indep_only', 'indep_only'),
                                   )
        if not self.arch_only and not self.indep_only:
            self.arch_only = self.indep_only = True
            
    def build(self):
        if not self.skip_build:
            if self.distribution.has_pure_modules() and self.indep_only:
                self.run_command('build_py')
            if self.distribution.has_ext_modules() and self.arch_only:
                self.run_command('build_ext')
                
                
class install (distutils.command.install.install):
    user_options = distutils.command.install.install.user_options + [
            ('data-dir=', 't', 'Directory where the application will find the data'),
            ('arch-only', None, 'Install only architecture dependent files'),
            ('indep-only', None, 'Install only architecture independent files'),
        ]
    boolean_options = ['arch-only', 'indep-only']
        
    def initialize_options(self):
        distutils.command.install.install.initialize_options(self)
        self.data_dir = None
        self.arch_only = False
        self.indep_only = False
        
    def finalize_options(self):
        distutils.command.install.install.finalize_options(self)
        if self.data_dir is None:
            self.data_dir = os.path.join(self.install_data, 'share')
        if not self.arch_only and not self.indep_only:
            self.arch_only = self.indep_only = True
        if not self.indep_only:
            self.__class__.sub_commands = [(cmd, func)
                            for cmd, func in distutils.command.install.install.sub_commands
                                if cmd == 'install_lib']
            
    def run(self):
        if not self.skip_build:
            # distutils.command.install.install.run() will run build, but we need
            # to modify a file between build and install
            build_cmd = self.distribution.get_command_obj('build')
            build_cmd.arch_only = self.arch_only
            build_cmd.indep_only = self.indep_only
            self.run_command('build')
            self.skip_build = True
            
        filename = os.path.join(self.build_lib, 'pybiklib', 'config.py')
        app_data_dir = os.path.join(self.data_dir, 'pybik')
        if self.indep_only:
            with open(filename, 'rt', encoding='utf-8') as f:
                text = f.read()
            for pattern, repl in [
                        (r'^(data_dir\s*=\s*).*$',      r'\1' + repr(self.data_dir)),
                        (r'^(appdata_dir\s*=\s*).*$',   r'\1' + repr(app_data_dir)),]:
                text = re.sub(pattern, repl, text, count=1, flags=re.MULTILINE)
            with open(filename, 'wt', encoding='utf-8') as f:
                f.write(text)
            
        distutils.command.install.install.run(self)
        if self.record:
            self.warn('The --record option is broken, files from the build_ui command are not recorded.')
        
        
class clean (distutils.command.clean.clean):
    user_options = distutils.command.clean.clean.user_options + [
            ('inplace', 'i',         'clean up files in the source directory'),
        ]
    boolean_options = distutils.command.clean.clean.boolean_options + [
            'inplace',
        ]
        
    def initialize_options(self):
        distutils.command.clean.clean.initialize_options(self)
        self.inplace = None
        
    def finalize_options(self):
        distutils.command.clean.clean.finalize_options(self)
        self.set_undefined_options('build',
                                       ('inplace', 'inplace'),
                                   )
        
    def run(self):
        def remove_tree_(directory):
            if os.path.exists(directory):
                remove_tree(directory, dry_run=self.dry_run, verbose=self.verbose)
            else:
                warn("%r does not exist -- can't clean it", directory)
        def remove_file(filename):
            if os.path.exists(filename):
                info('removing %r', filename)
                if not self.dry_run:
                    os.remove(filename)
            else:
                warn("%r does not exist -- can't clean it", filename)
                
        if self.all:
            for _dir in ['mo', 'man', 'models']:
                remove_tree_(os.path.join(self.build_base, _dir))
        distutils.command.clean.clean.run(self)
        if self.inplace:
            for dirname in ('pybiklib/', 'pybiklib/ui/', 'data/plugins/', 'pybiktest/'):
                for filename in glob(dirname + '*.pyc'):
                    remove_file(filename)
                remove_tree_(dirname + '__pycache__')
            if self.all:
                dirnames = ['pybiklib/ui', 'data/locale', 'data/models']
                for dirname in dirnames:
                    remove_tree_(dirname)
                for filename in glob('pybiklib/*.so'):
                    remove_file(filename)
                remove_file('pybik')
                
                
class sdist(distutils.command.sdist.sdist):
    user_options = distutils.command.sdist.sdist.user_options + [
            ('debian-names', None,  'Create archive files with Debian names'),
        ]
    boolean_options = distutils.command.sdist.sdist.boolean_options + [
            'debian-names'
        ]
        
    def initialize_options(self):
        distutils.command.sdist.sdist.initialize_options(self)
        self.debian_names = False
        
    def run(self):
        from tools import create_docs
        copyright = create_docs.read_utf8('debian/copyright')   # pylint: disable=W0622
        copyright = copyright.replace(
                ' The full text of the GPL is distributed in\n'
                ' /usr/share/common-licenses/GPL-3 on Debian systems.',
                ' The full text of the GPL is distributed in\n'
                ' the original source archive in the file COPYING.')
        create_docs.write_utf8('copyright', copyright)
        create_docs.create_README()
        
        def remove_tree_(directory):
            if os.path.exists(directory):
                remove_tree(directory, dry_run=self.dry_run, verbose=self.verbose)
            else:
                warn("%r does not exist -- can't clean it", directory)
        remove_tree_('csrc')
        self.run_command('build_csrc')
        distutils.command.sdist.sdist.run(self)
        for archive_file in self.archive_files:
            if self.debian_names:
                debian_file = archive_file.replace('-', '_', 1).replace('.tar', '.orig.tar', 1)
                os.rename(archive_file, debian_file)
        remove_tree_('csrc')
        
    def get_file_list(self):
        self.filelist.append('copyright')
        for f in glob('po/*.po'):
            if po_isempty(f):
                print('skipping empty po file', f)
            else:
                self.filelist.files.append(f)
        distutils.command.sdist.sdist.get_file_list(self)
        
        
setup(
        # Metadata
        name=config.PACKAGE,
        version=config.VERSION,
        author=config.AUTHOR,
        author_email=config.CONTACT_EMAIL,
        url=config.WEBSITE,
        description=config.SHORT_DESCRIPTION,
        long_description=config.LONG_DESCRIPTION,
        download_url=config.DOWNLOADS,
        classifiers=[
                'Development Status :: 5 - Production/Stable',
                'Environment :: X11 Applications :: Qt',
                'Intended Audience :: End Users/Desktop',
                'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
                'Operating System :: POSIX',
                'Operating System :: POSIX :: Linux',
                'Programming Language :: Python :: 3',
                'Programming Language :: Cython',
                'Programming Language :: OpenGL Shading Language',
                'Topic :: Games/Entertainment :: Puzzle Games',
            ],
        license=config.LICENSE_NAME,
        # Metadata added in overloaded Distribution
        bug_contact=config.CONTACT_FILEBUG,
        
        # Files
        scripts=['pybiklib/pybik.py'],
        data_files=[
                        ('share/applications', ['data/applications/pybik.desktop']),
                        ('share/pixmaps', ['data/pixmaps/pybik.png']),
                        ('share/pybik/ui', glob('data/ui/*.*')),
                        ('share/pybik/ui/images', glob('data/ui/images/*')),
                        ('share/pybik/plugins', glob('data/plugins/*.py')),
                        ('share/pybik/plugins', glob('data/plugins/*.algorithm')),
                        ('share/pybik/shaders', glob('data/shaders/*')),
                        ('share/pybik/tests', glob('data/tests/*')),
                        ('share/pybik/', ['data/GPL-3']),
                    ],
        qt_ui_files=[('pybiklib/ui', glob('data/ui/*.ui'))],
        ext_modules=[
                        Extension('pybiklib/_gldraw',
                                    ['pybiklib/gldraw.py'],
                                    language='c++',
                                    depends=['pybiklib/gl.pxd',
                                             'pybiklib/glm.pxd'],
                                    libraries=['GL'],
                                    define_macros=[('GL_GLEXT_PROTOTYPES', None)]),
                        Extension('pybiklib/_glarea',
                                    ["pybiklib/glarea.py"],
                                    language='c++',
                                    depends=['pybiklib/gl.pxd',
                                             'pybiklib/glm.pxd',
                                             'pybiklib/gldraw.py'],
                                    libraries=['GL'],
                                    define_macros=[('GL_GLEXT_PROTOTYPES', None)]),
                    ],
        packages=['pybiklib', 'pybiktest'],
        package_dir={'pybiklib': 'pybiklib', 'pybiktest': 'pybiktest'},
        package_data={'pybiktest': ['data/*']},
        
        # setup classes
        distclass=Distribution,
        cmdclass={
                    'build': build,
                    'build_csrc': build_csrc,
                    'build_ext': build_ext,
                    'build_scripts': build_scripts,
                    'build_ui': build_ui,
                    'build_models': build_models,
                    'build_i18n': build_i18n,
                    'build_man': build_man,
                    'install_data': distutils.command.install_data.install_data,
                    'install_lib': install_lib,
                    'install': install,
                    'clean': clean,
                    'sdist': sdist,
                },
     )
     



#-*- coding:utf-8 -*-

#  Pybik -- A 3 dimensional magic cube game.
#  Copyright © 2009-2013  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Ported from GNUbik
# Original filename: glarea.c
# Original copyright and license: 2003, 2004  John Darrington,  GPL3+


import os
from collections import namedtuple
import math
from time import sleep
from contextlib import contextmanager

# pylint: disable=W0614,W0401
from .debug import *

from PyQt4.QtOpenGL import *
from PyQt4.QtCore import *
from PyQt4.QtCore import pyqtSignal as Signal
from PyQt4.QtGui import *
# pylint: enable=W0614,W0401

# no need for OpenGL module for just two constants
GL_RGBA, GL_SRGB_ALPHA, GL_TEXTURE_2D = 6408, 35906, 3553
if DEBUG:
    try:
        from OpenGL import GL
    except Exception:
        pass
    else:
        assert GL.GL_RGBA == GL_RGBA
        assert GL.GL_SRGB_ALPHA == GL_SRGB_ALPHA
        assert GL.GL_TEXTURE_2D == GL_TEXTURE_2D

from . import config
from .textures import textures
from .settings import settings
from . import model


class Design:
    def __init__(self):
        self.bevel_color = (15, 15, 15)
        self.label_colors = [(0, 0, 0) for i in range(6)]
        self.imagemodes = [0] * 6
        self.texnames = [-1] * 6
        
        
class ShaderInterface:
    shader_prefix = b'''
#version 120
#line 0
'''
    
    def init_shaders(self, shadername):
        if shadername is None:
            self.load_shader()
        else:
            self.set_shader(shadername)
        if DEBUG_DRAW:
            debug('Creating "hud" shaders:')
            vertex_source = self._read_shader('hud.vert')
            fragment_source = self._read_shader('pick.frag')
            self.glarea.gl_create_hud_program(vertex_source, fragment_source)
        debug('Creating "pick" shaders:')
        vertex_source = self._read_shader('pick.vert')
        fragment_source = self._read_shader('pick.frag')
        self.glarea.gl_create_pick_program(vertex_source, fragment_source)
        
    @staticmethod
    def _read_shader(filename):
        filename = os.path.join(config.appdata_dir, 'shaders', filename)
        with open(filename, 'rb') as sfile:
            source = sfile.read()
        return ShaderInterface.shader_prefix + source
        
    def set_shader(self, shadername):
        self.shadername = shadername
        debug('Creating "%s" shaders:' % shadername)
        vertex_source = self._read_shader(shadername + '.vert')
        fragment_source = self._read_shader(shadername + '.frag')
        self.glarea.gl_create_render_program(vertex_source, fragment_source)
        
    def load_shader(self):
        shadername = 'lighting' if settings.draw.lighting else 'simple'
        self.set_shader(shadername)
        
        
class CubeAreaBase (QGLWidget, ShaderInterface):
    animation_ending = Signal(bool)
    request_rotation = Signal((int, int, bool))
    request_swap_blocks = Signal((int, int, int, bool))
    request_rotate_block = Signal((int, bool))
    drop_color = Signal((int, str, str))
    drop_file = Signal((int, str, str))
    debug_text = Signal(str)
    
    def __init__(self, opts):
        glformat = QGLFormat()
        if settings.draw.samples > 0:
            glformat.setSampleBuffers(True)
            glformat.setSamples(2**settings.draw.samples)
        QGLWidget.__init__(self, glformat)
        if QGLFormat.openGLVersionFlags() & QGLFormat.OpenGL_Version_2_1:
            self.texture_format = GL_SRGB_ALPHA
        else:
            #FIXME: With OpenGL version < 2.1 images are not gamma correct
            self.texture_format = GL_RGBA
        
        self.shadername = opts.shader
        self.design = Design()
        self.model = model.empty_model
        
        self.last_mouse_x = -1
        self.last_mouse_y = -1
        self.button_down_background = False
        self.mouse_xy = -1, -1
        # Structure to hold copy of the last selection taken or None
        self.pickdata = None
        self.editing_model = False
        settings.keystore.changed.connect(self.on_settings_changed, Qt.QueuedConnection)
        
        if DEBUG_FPS:
            self.monotonic_time = QElapsedTimer()
            self.monotonic_time.start()
            self.render_count = 0
            self.fps = 0.
            
        self.speed = settings.draw.speed
        
        # Initialize the render engine
        self.gldraw, self.glarea = self.import_gl_engine()
        self.glarea.init_module()
        self.gldraw.init_module()
        self.rotation_x, self.rotation_y = self.glarea.set_rotation_xy(*settings.draw.default_rotation)
        self.glarea.set_antialiasing(self.format().sampleBuffers())
        self.init_design()
        
        self.setAcceptDrops(True)
        self.setFocusPolicy(Qt.StrongFocus)
        self.setMinimumSize(200, 200)
        
        self.cursors = None
        self.load_cursors()
        self.update_selection_pending = False
        self.timer_update_selection = QTimer(self)
        self.timer_update_selection.setSingleShot(True)
        self.timer_update_selection.timeout.connect(self.on_idle_update_selection)
        self.set_cursor()
        self.setMouseTracking(True)
        
    def import_gl_engine(self):
        if not DEBUG_PUREPYTHON:
            from . import _gldraw as gldraw
            from . import _glarea as glarea
        else:
            try:
                import OpenGL.GL
            except ImportError as e:
                print('The pure Python mode needs PyOpenGL (for Python 3):', e)
                raise SystemExit(1)
            from . import gldraw
            from . import glarea
        return gldraw, glarea
        
    def start(self):
        pass
        
    def stop(self):
        self.glarea.gl_exit()
        self.delete_textures()
        # this line prevents a segmentation fault with PySide if game->quit selected
        self.setMouseTracking(False)
        settings.keystore.changed.disconnect(self.on_settings_changed)
        self.timer_update_selection.timeout.disconnect(self.on_idle_update_selection)
        
    def load_cursors(self):
        cursors = []
        # Load 3 cursors from file (n - ne)
        for i, (x, y) in enumerate([(8, 0), (15, 0), (15, 0)]):
            filename = os.path.join(config.UI_DIR, 'mouse_{}.png'.format(i))
            image = QImage(filename)
            cursors.append((image, x, y))
        # 1 cursor (nnw)
        image, x, y = cursors[1]
        cursors.insert(0, (image.mirrored(True, False), 15-x, y))
        # 12 cursors (ene - nw)
        transform = QTransform()
        transform.rotate(90)
        for i in range(4, 16):
            image, x, y = cursors[-4]
            cursors.append((image.transformed(transform), 15-y, x))
        cursors.append(cursors[0])
        self.cursors = [QCursor(QPixmap.fromImage(image), x, y) for image, x, y in cursors[1:]]
        # cursor for center faces
        filename = os.path.join(config.UI_DIR, 'mouse_ccw.png')
        cursor = QCursor(QPixmap(filename), 7, 7)
        self.cursors.append(cursor)
        # background cursor
        cursor = QCursor()
        cursor.setShape(Qt.CrossCursor)
        self.cursors.append(cursor)
        
    def init_design(self):
        rgba = QColor()
        for i in range(6):
            rgba.setNamedColor(settings.theme.face[i].color)
            self.design.label_colors[i] = rgba.red(), rgba.green(), rgba.blue()
            self.design.imagemodes[i] = settings.theme.face[i].mode
        self.set_background_color(settings.theme.bgcolor)
        
    stock_texnames = [0]
    def set_face_texture(self, faceidx, imagefile):
        if imagefile.startswith('/'):
            pixbuf = textures.create_pixbuf_from_file(imagefile)
        else:
            pixbuf = textures.get_stock_pixbuf(imagefile)
        texname = self.bindTexture(pixbuf, GL_TEXTURE_2D, self.texture_format,
                                   QGLContext.LinearFilteringBindOption | QGLContext.MipmapBindOption)
        if not imagefile.startswith('/') and texname not in self.stock_texnames:
            self.stock_texnames.append(texname)
        if self.design.texnames[faceidx] not in self.stock_texnames:
            self.deleteTexture(self.design.texnames[faceidx])
        self.design.texnames[faceidx] = texname
        
    def delete_textures(self):
        for texname in self.design.texnames:
            if texname not in self.stock_texnames:
                self.deleteTexture(texname)
        del self.design.texnames[:]
        for texname in self.stock_texnames:
            if texname:
                self.deleteTexture(texname)
        self.stock_texnames[:] = [0]
        
    def set_model(self, model_, blocks=None):
        self.model = model_
        glmodeldata = self.get_glmodeldata(settings.draw.selection)
        with self.lock_glcontext():
            self.glarea.set_frustrum(self.model.bounding_sphere_radius, settings.draw.zoom)
            self.gldraw.gl_set_data(*glmodeldata)
            if blocks is not None:
                self.gldraw.set_transformations(blocks)
        self.update_selection()
        self.update()
        
    def get_glmodeldata(self, selection_mode):
        (   glvertices, glnormals, glcolors, gltexpos,
            labelinfos,
            idx_vbevel, cnts_block,
            idx_pick, cnt_pick,
            idx_debug,
        ) = self.model.gl_vertex_data(selection_mode, self.design)
        ptrnormals = len(glvertices)
        ptrcolors = ptrnormals + len(glnormals)
        ptrtexpos = ptrcolors + len(glcolors)
        return (len(self.model.blocks),
                glvertices + glnormals + glcolors + gltexpos,
                (ptrnormals, ptrcolors, ptrtexpos),
                (labelinfos, idx_vbevel, cnts_block, idx_pick, cnt_pick, idx_debug),
                self.design.texnames)
                
    def set_glmodel(self, selection_mode):
        glmodeldata = self.get_glmodeldata(selection_mode)
        with self.lock_glcontext():
            self.gldraw.gl_set_data(*glmodeldata)
        
    def set_transformations(self, cubestate):
        with self.lock():
            self.gldraw.set_transformations(cubestate.blocks)
        
    def do_initializeGL(self):
        if DEBUG_MSG:
            glcontext = self.context()
            glformat = glcontext.format()
            glrformat = glcontext.requestedFormat()
            def printglattr(name, *attrs):
                print('  {}: '.format(name), end='')
                def get_value(glformat, attr):
                    if isinstance(attr, str):
                        return getattr(glformat, attr)()
                    else:
                        return attr(glformat)
                values = [get_value(glformat, a) for a in attrs]
                rvalues = [get_value(glrformat, a) for a in attrs]
                if values == rvalues:
                    print(*values)
                else:
                    print(*values, end=' (')
                    print(*rvalues, end=')\n')
            print('OpenGL format:')
            printglattr('accum', 'accum', 'accumBufferSize')
            printglattr('alpha', 'alpha', 'alphaBufferSize')
            printglattr('rgb', 'redBufferSize', 'greenBufferSize', 'blueBufferSize')
            printglattr('depth', 'depth', 'depthBufferSize')
            printglattr('directRendering', 'directRendering')
            printglattr('doubleBuffer', 'doubleBuffer')
            printglattr('hasOverlay', 'hasOverlay')
            printglattr('version', lambda glformat: '{}.{} 0x{:x}'.format(
                                                            glformat.majorVersion(),
                                                            glformat.minorVersion(),
                                                            int(glformat.openGLVersionFlags())))
            printglattr('plane', 'plane')
            printglattr('profile', 'profile')
            printglattr('rgba', 'rgba')
            printglattr('samples', 'sampleBuffers', 'samples')
            printglattr('stencil', 'stencil', 'stencilBufferSize')
            printglattr('stereo', 'stereo')
            printglattr('swapInterval', 'swapInterval')
        self.init_shaders(self.shadername)
        for i in range(6):
            self.set_face_texture(i, settings.theme.face[i].image)
        self.glarea.gl_init()
        
    def do_paintGL(self):
        self.glarea.gl_render()
        if DEBUG:
            if DEBUG_DRAW:
                if self.pickdata and not self.pickdata.center:
                    self.glarea.gl_render_select_debug()
            if DEBUG_FPS:
                self.render_count += 1
                if self.monotonic_time.elapsed() > 1000:
                    elapsed = self.monotonic_time.restart()
                    self.fps = 1000. / elapsed * self.render_count
                    self.render_count = 0
                    self.debug_text.emit("FPS %.1f" % self.fps)
        
    def do_resizeGL(self, width, height):
        self.glarea.gl_resize(width, height)
        
    MODIFIER_MASK = int(Qt.ShiftModifier | Qt.ControlModifier)
    def keyPressEvent(self, event):
        modifiers = int(event.modifiers()) & self.MODIFIER_MASK
        if modifiers:
            return QGLWidget.keyPressEvent(self, event)
        elif event.key() == Qt.Key_Right:
            self.rotation_x += 2
        elif event.key() == Qt.Key_Left:
            self.rotation_x -= 2
        elif event.key() == Qt.Key_Up:
            self.rotation_y -= 2
        elif event.key() == Qt.Key_Down:
            self.rotation_y += 2
        else:
            return QGLWidget.keyPressEvent(self, event)
            
        self.rotation_x, self.rotation_y = self.glarea.set_rotation_xy(self.rotation_x, self.rotation_y)
        self.update()
        self.update_selection()
        
    PickData = namedtuple('PickData', 'maxis mslice mdir block symbol face center angle')
    def pick_polygons(self, x, y):
        '''Identify the block at screen co-ordinates x,y.'''
        
        height = self.height()
        with self.lock_glcontext():
            index = self.glarea.gl_pick_polygons(x, height-y)
        if not index:
            self.pickdata = None
            return
        maxis, mslice, mdir, face, center, block, symbol, face_center, edge_center = self.model.pick_data[index]
        
        if center:
            angle = None
        else:
            angle = self.glarea.get_cursor_angle(face_center, edge_center)
        self.pickdata = self.PickData(maxis, mslice, mdir, block, symbol, face, center, angle)
        if DEBUG_DRAW:
            text = ("block {}, face {} ({}), center {}\n"
                    "axis {}, slice {}, dir {}\n"
                    "angle {}\n"
                    "mouse {}, {}").format(
                    block.index, symbol, face, center,
                    maxis, mslice, mdir,
                    angle, *self.mouse_xy)
            self.debug_text.emit(text)
        
    def update_selection(self):
        '''This func determines which block the mouse is pointing at'''
        if self.animation_active:
            if self.pickdata is not None:
                self.pickdata = None
                self.set_cursor()
            return
        if self.update_selection_pending:
            return
        self.update_selection_pending = True
        self.timer_update_selection.start()
        
    def on_idle_update_selection(self):
        if self.animation_active:
            if self.pickdata is not None:
                self.pickdata = None
                self.set_cursor()
        else:
            self.pick_polygons(*self.mouse_xy)
            self.set_cursor()
            if DEBUG_DRAW:
                self.update()
        self.update_selection_pending = False
        
    def mouseMoveEvent(self, event):
        self.mouse_xy = event.x(), event.y()
        
        if not self.button_down_background:
            self.update_selection()
            return
            
        # perform rotation
        offset_x = event.x() - self.last_mouse_x
        offset_y = event.y() - self.last_mouse_y
        self.rotation_x, self.rotation_y = self.glarea.set_rotation_xy(
                                                        self.rotation_x + offset_x,
                                                        self.rotation_y + offset_y)
        self.rotation_x -= offset_x
        self.rotation_y -= offset_y
        self.update()
        
    def mousePressEvent(self, event):
        if self.pickdata is not None:
            if self.animation_active:
                return
            # make a move
            if self.editing_model:
                if event.modifiers() & Qt.ControlModifier:
                    if event.button() == Qt.LeftButton:
                        self.request_rotate_block.emit(self.pickdata.block.index, False)
                    elif event.button() == Qt.RightButton:
                        self.request_rotate_block.emit(self.pickdata.block.index, True)
                else:
                    if event.button() == Qt.LeftButton:
                        self.request_swap_blocks.emit(self.pickdata.block.index,
                                    self.pickdata.maxis, self.pickdata.mslice, self.pickdata.mdir)
            else:
                mslice = -1 if event.modifiers() & Qt.ControlModifier else self.pickdata.mslice
                if event.button() == Qt.LeftButton:
                    self.request_rotation.emit(self.pickdata.maxis, mslice, self.pickdata.mdir)
                elif event.button() == Qt.RightButton and settings.draw.selection_nick == 'simple':
                    self.request_rotation.emit(self.pickdata.maxis, mslice, not self.pickdata.mdir)
        elif event.button() == Qt.LeftButton:
            # start rotation
            self.button_down_background = True
            self.last_mouse_x = event.x()
            self.last_mouse_y = event.y()
        self.update()
        
    def mouseReleaseEvent(self, event):
        if event.button() != Qt.LeftButton:
            return
            
        if self.button_down_background:
            # end rotation
            self.rotation_x += event.x() - self.last_mouse_x
            self.rotation_y += event.y() - self.last_mouse_y
            self.button_down_background = False
            self.update_selection()
            
    def wheelEvent(self, event):
        if event.orientation() == Qt.Vertical:
            zoom = settings.draw.zoom * math.pow(1.1, -event.delta() / 120)
            zoom_min, zoom_max = settings.draw.zoom_range
            if zoom < zoom_min:
                zoom = zoom_min
            if zoom > zoom_max:
                zoom = zoom_max
            settings.draw.zoom = zoom
            
    def dragEnterEvent(self, event):
        debug('drag enter:', event.mimeData().formats())
        if (event.mimeData().hasFormat("text/uri-list") or
                event.mimeData().hasFormat("application/x-color")):
            event.acceptProposedAction()
            
    def dropEvent(self, event):
        # when a drag is in progress the pickdata is not updated, so do it now
        self.pick_polygons(event.pos().x(), event.pos().y())
        
        mime_data = event.mimeData()
        if mime_data.hasFormat("application/x-color"):
            color = mime_data.colorData()
            if color is None:
                debug("Received invalid color data")
                return
                
            if self.pickdata is not None:
                self.drop_color.emit(self.pickdata.block.index, self.pickdata.symbol, color.name())
            else:
                self.drop_color.emit(-1, '', color.name())
        elif mime_data.hasFormat("text/uri-list"):
            if self.pickdata is None:
                debug('Background image is not supported.')
                return
            uris = mime_data.urls()
            for uri in uris:
                if not uri.isLocalFile():
                    debug('filename "%s" not found or not a local file.' % uri.toString())
                    continue
                filename = uri.toLocalFile()
                if not filename or not os.path.exists(filename):
                    debug('filename "%s" not found.' % filename)
                    continue
                self.drop_file.emit(self.pickdata.block.index, self.pickdata.symbol, filename)
                break # For now,  just use the first one.
        # Ignore all others
        
    def set_cursor(self):
        if self.pickdata is None or self.button_down_background:
            index = -1
        elif self.pickdata.angle is None:
            index = -2
        else:
            index = int((self.pickdata.angle+180) / 22.5 + 0.5) % 16
        self.setCursor(self.cursors[index])
        
    def set_std_cursor(self):
        QTimer.singleShot(0, lambda: self.unsetCursor())
        
    def set_editing_mode(self, enable):
        self.editing_model = enable
        self.set_glmodel(0 if enable else settings.draw.selection)
        self.update_selection()
        
    def set_background_color(self, color):
        rgba = QColor()
        rgba.setNamedColor(color)
        self.glarea.set_background_color(rgba.redF(), rgba.greenF(), rgba.blueF())
        
    def reset_rotation(self):
        '''Reset cube rotation'''
        self.rotation_x, self.rotation_y = self.glarea.set_rotation_xy(*settings.draw.default_rotation)
        self.update()
        
    def reload_shader(self):
        with self.lock_glcontext():
            self.set_shader(self.shadername)
        self.update()
        
    def on_settings_changed(self, key):
        if key == 'draw.speed':
            self.speed = settings.draw.speed
        elif key == 'draw.lighting':
            with self.lock_glcontext():
                self.load_shader()
        elif key == 'draw.samples':
            if self.format().samples():
                samples = 2**settings.draw.samples
                if samples == 1:
                    self.glarea.set_antialiasing(False)
                elif samples == self.format().samples():
                    self.glarea.set_antialiasing(True)
        elif key == 'draw.selection':
            self.set_glmodel(settings.draw.selection)
            self.update_selection()
        elif key.startswith('theme.face.'):
            i = int(key.split('.')[2])
            if key == 'theme.face.{}.color'.format(i):
                color = QColor()
                color.setNamedColor(settings.theme.face[i].color)
                self.design.label_colors[i] = (color.red(), color.green(), color.blue())
                self.set_glmodel(settings.draw.selection)
            elif key == 'theme.face.{}.image'.format(i):
                with self.lock_glcontext():
                    self.set_face_texture(i, settings.theme.face[i].image)
                self.set_glmodel(settings.draw.selection)
            elif key == 'theme.face.{}.mode'.format(i):
                self.design.imagemodes[i] = settings.theme.face[i].mode
                self.set_glmodel(settings.draw.selection)
        elif key == 'theme.bgcolor':
            self.set_background_color(settings.theme.bgcolor)
        elif key == 'draw.zoom':
            self.glarea.set_frustrum(self.model.bounding_sphere_radius, settings.draw.zoom)
        else:
            debug('Unknown settings key changed:', key)
        self.update()
        
        
class AnimationBaseInterface:
    def __init__(self):
        self.stop_requested = False
        self.animation_active = False
        
    def animate_rotation(self, move_data, blocks, stop_after):
        self.stop_requested = stop_after
        axis = (self.model.axesI if move_data.dir else self.model.axes)[move_data.axis]
        angle = self.model.symmetry[move_data.axis]
        self.gldraw.set_animation_start(blocks, angle, *axis)
        self.animation_active = True
            
    def animation_end(self):
        self.animation_active = False
        
    def _on_animate(self):
        increment = self.speed * 1e-02 * 20
        increment = min(increment, 45)
        unfinished = self.gldraw.set_animation_next(increment)
        if unfinished:
            return True
            
        # we have finished the animation sequence now
        return False
        
    def animate_abort(self, update=True):
        if update:
            self.animation_ending.emit(self.stop_requested)
            self.update()
            self.update_selection()
        else:
            self.animation_end()
        self.stop_requested = False
        
        
class AnimationInterface (AnimationBaseInterface):
    def __init__(self):
        AnimationBaseInterface.__init__(self)
        self.timer_animate = QTimer() # the animate timer
        self.timer_animate.timeout.connect(self._on_animate)
        
    def animate_rotation(self, move_data, blocks, stop_after):
        AnimationBaseInterface.animate_rotation(self, move_data, blocks, stop_after)
        self.timer_animate.start(0 if DEBUG_VFPS else 20)
        self.update_selection()
        
    def _on_animate(self):
        if AnimationBaseInterface._on_animate(self):
            self.update()
            return
            
        # we have finished the animation sequence now
        self.animation_ending.emit(self.stop_requested)
        self.update()
        self.update_selection()
        self.stop_requested = False
        
    def animation_end(self):
        AnimationBaseInterface.animation_end(self)
        self.timer_animate.stop()
        
        
class CubeArea (CubeAreaBase, AnimationInterface):
    def __init__(self, opts):
        CubeAreaBase.__init__(self, opts)
        AnimationInterface.__init__(self)
        
    def stop(self):
        CubeAreaBase.stop(self)
        self.timer_animate.timeout.disconnect(self._on_animate)
        
    initializeGL = CubeAreaBase.do_initializeGL
    resizeGL = CubeAreaBase.do_resizeGL
    paintGL = CubeAreaBase.do_paintGL
    
    @contextmanager
    def lock(self):
        yield
        
    lock_glcontext = lock
    
    
class AnimationThreadedInterface (AnimationBaseInterface):
    def animate_rotation(self, move_data, blocks, stop_after):
        with self.lock():
            AnimationBaseInterface.animate_rotation(self, move_data, blocks, stop_after)
        self.render_condition.wakeAll()
        self.update_selection()
        
    def _on_animate(self):
        if AnimationBaseInterface._on_animate(self):
            return
            
        # we have finished the animation sequence now
        self.animation_active = False
        self.animation_ending.emit(self.stop_requested)
        self.update_selection()
        self.stop_requested = False
        
        
class CubeAreaThreaded (CubeAreaBase, AnimationThreadedInterface):
    def __init__(self, opts):
        CubeAreaBase.__init__(self, opts)
        AnimationThreadedInterface.__init__(self)
        
        self.render_thread = RenderThread(self.thread_loop)
        
        self.request_render = True
        self.request_resize = False
        self.render_mutex = QMutex()
        self.render_condition = QWaitCondition()
        # Buffer swap is handled in the rendering thread
        self.setAutoBufferSwap(False)
        
    def start(self):
        self.render_thread.start()
        self.render_condition.wakeAll()
        
    def stop(self):
        '''Performs a save shutdown of the render thread'''
        self.request_render = False
        self.render_condition.wakeAll()
        self.render_thread.wait()
        CubeAreaBase.stop(self)
        
    @contextmanager
    def lock(self):
        self.render_mutex.lock()
        yield
        self.render_mutex.unlock()
        
    @contextmanager
    def lock_glcontext(self):
        self.render_mutex.lock()
        self.makeCurrent()
        yield
        self.doneCurrent()
        self.render_mutex.unlock()
        
    @contextmanager
    def release_glcontext(self):
        self.doneCurrent()
        self.render_mutex.unlock()
        yield
        self.render_mutex.lock()
        self.makeCurrent()
        
    def paintEvent(self, event):
        self.render_condition.wakeAll()
        
    def resizeEvent(self, event):
        self.requested_size = event.size()
        self.request_resize = True
        self.render_condition.wakeAll()
        
    def thread_loop(self):
        with self.lock_glcontext():
            self._thread_loop()
            
    def _thread_loop(self):
        self.do_initializeGL()
        while self.request_render:
            if self.request_resize:
                self.do_resizeGL(self.requested_size.width(), self.requested_size.height())
                self.request_resize = False
            self.do_paintGL()
            self.swapBuffers()
            if self.animation_active:
                with self.release_glcontext():
                    if not DEBUG_VFPS:
                        sleep(.02)
                    self._on_animate()
            else:
                self.doneCurrent()
                self.render_condition.wait(self.render_mutex)
                self.makeCurrent()
        
        
class RenderThread (QThread):
    def __init__(self, func):
        QThread.__init__(self)
        self.func = func
        
    def run(self):
        self.func()
        
        




# generated with: ./merge-translators.sh

translators = (
[
  ('Asturian', 'ast', [
    ('Xuacu Saturio', 'https://launchpad.net/~xuacusk8'),
    ('ivarela', 'https://launchpad.net/~ivarela'),
  ]),
  ('Bengali', 'bn', [
    ('Iftekhar Mohammad', 'https://launchpad.net/~iftekhar'),
  ]),
  ('Bosnian', 'bs', [
    ('Kenan Dervišević', 'https://launchpad.net/~kenan3008'),
  ]),
  ('Brazilian Portuguese', 'pt_BR', [
    ('Fábio Nogueira', 'https://launchpad.net/~fnogueira'),
    ('Rafael Neri', 'https://launchpad.net/~rafepel'),
  ]),
  ('Chinese (Simplified)', 'zh_CN', [
    ('Xiaoxing Ye', 'https://launchpad.net/~xiaoxing'),
  ]),
  ('Chinese (Traditional)', 'zh_TW', [
    ('Po-Chun Huang', 'https://launchpad.net/~aphroteus'),
  ]),
  ('English (United Kingdom)', 'en_GB', [
    ('Andi Chandler', 'https://launchpad.net/~bing'),
    ('Anthony Harrington', 'https://launchpad.net/~untaintableangel'),
    ('B Clausius', 'https://launchpad.net/~barcc'),
  ]),
  ('French', 'fr', [
    ('Aurélien Ribeiro', 'https://launchpad.net/~aurel-koala'),
    ('Baptiste Fontaine', 'https://launchpad.net/~bfontaine'),
    ('Célestin Taramarcaz', 'https://launchpad.net/~celestin'),
    ('Havok Novak', 'https://launchpad.net/~havok-novak'),
    ('Nicolas Delvaux', 'https://launchpad.net/~malizor'),
    ('Sylvie Gallet', 'https://launchpad.net/~sylvie-gallet'),
    ('lann', 'https://launchpad.net/~lann'),
  ]),
  ('Galician', 'gl', [
    ('Miguel Anxo Bouzada', 'https://launchpad.net/~mbouzada'),
  ]),
  ('German', 'de', [
    ('B Clausius', 'https://launchpad.net/~barcc'),
  ]),
  ('Greek', 'el', [
    ('George Christofis', 'https://launchpad.net/~geochr'),
    ('mara sdr', 'https://launchpad.net/~paren8esis'),
  ]),
  ('Hebrew', 'he', [
    ('Yaron', 'https://launchpad.net/~sh-yaron'),
  ]),
  ('Indonesian', 'id', [
    ('Trisno Pamuji', 'https://launchpad.net/~tri.snowman'),
  ]),
  ('Italian', 'it', [
    ('Alfio Missaglia', 'https://launchpad.net/~missaglialfio'),
    ('Claudio Arseni', 'https://launchpad.net/~claudio.arseni'),
    ('Francesco Muriana', 'https://launchpad.net/~f-muriana'),
  ]),
  ('Japanese', 'ja', [
    ('epii', 'https://launchpad.net/~epii'),
  ]),
  ('Kirghiz', 'ky', [
    ('ballpen', 'https://launchpad.net/~ballpen'),
  ]),
  ('Malay', 'ms', [
    ('abuyop', 'https://launchpad.net/~abuyop'),
  ]),
  ('Polish', 'pl', [
    ('Mike08', 'https://launchpad.net/~micou8'),
    ('Szymon Nieznański', 'https://launchpad.net/~isamu715'),
  ]),
  ('Russian', 'ru', [
    ('Oleg Koptev', 'https://launchpad.net/~koptev-oleg'),
    ('scientistnik', 'https://launchpad.net/~nozdrin-plotnitsky'),
  ]),
  ('Serbian', 'sr', [
    ('Мирослав Николић', 'https://launchpad.net/~lipek'),
  ]),
  ('Spanish', 'es', [
    ('Adolfo Jayme Barrientos', 'https://launchpad.net/~fitoschido'),
    ('Dante Díaz', 'https://launchpad.net/~dante'),
    ('Eduardo Alberto Calvo', 'https://launchpad.net/~edu5800'),
  ]),
  ('Ukrainian', 'uk', [
    ('Yuri Chornoivan', 'https://launchpad.net/~yurchor-gmail'),
  ]),
  ('Uzbek', 'uz', [
    ('Akmal Xushvaqov', 'https://launchpad.net/~uzadmin'),
  ]),
]
)
